""" The swarm_util module provides a host of utilities function.

This module stores a vararity of different functions and constants,
which is deemed general usefull when working on swarm data,
specifically when such data is loaded from 
the viresclient python client.

There is a mix of general constants, such as the permuability of vacum,
and different types of functions.
The functions can generally be categorised into three types:
Higher order functions, functions for working on arrays 
of scalars or vectors, and functions to work with viresclient.
These categories have their own sections, with a short descriptions on each.

Higher Order Functions
----------------------
Higher order functions are functions that operates on other functions.
This section also contains utilities for higher other higher order functions.

call
    A function to call other functions.
    Its main use is propergating options down the call stack.
maybe_pops
    A utility for call, works kind of like an intersection,
    but pops (remove and return) key-value pairs from a dictionary.
maybe_keys
    A utility for call, works kind of like an intersection,
    but returns key-value pairs from a dictionary.

    
Array Utilities
---------------
These functions are used to work on numpy arrays of scalars or vectors.
They are mostly geared toward dealign with time-like series.

pack_3d
    Take 3 equaly long arrays of scalars and make an array of 3d vectors.
as_3d
    Like pack_3d, but uses dublicates of the same array of scalars.
    Usefull for aggregating scalar-vector operations over an array.
map_3d
    A map function over arrays of 3d vectors,
    where the function maped takes a 3d vector to a 3d vector.
NEC_to_VSC
    Build a transformation from the NEC frame to the VSC frame.
    The return is a function, which will transfrom
    its input from the NEC frame to the VSC frame.
delta
    Calculate the changes of a time-like series.
means
    Calculate the intermediate value (by linear interpolation)
    for a time-like series.
spherical_delta
    Calculate the changes in spherical coordinates as NEC frame vectors.
curl
    Calculate a single component of curl for an array of vectors.
inclination
    Calculate the inclination for an awway of vectors.
    
Viresclient
-----------
These are the functions related to fetching and loading data
from the viresclient python client.

request_data
    Builds and send a request for a CDF file to vires.
    Most usefull for defaulting all the relevant parameters.
read_cdf
    A function to read a cdf file based on a dictionary pairing
    of attributes you want to read from the cdf file,
    and what you want to refere to them as afterwards.
"""
# """ Utilities for working on swarm data,
# when using the viresclient python client.
# """

__version__ = '0.1.3'
__author__ = 'Ask Neve Gamby'

import datetime as date
import numpy as np
import cdflib

from viresclient import SwarmRequest

from . import safety as safe_user

MU_0 = 4.0 * np.pi * 10**(-7)
TO_RADIANS = np.pi / 180


def maybe_pops(source, keys):
    """ Create a new dict based on the intersection of source keys and keys..
    
    Create a new dictionary. Each pair of key and value comes from source.
    Inclusion in the result is based on intersection of source's keys and keys.
    
    Parameters
    ----------
    source : dict
        A dictionary forming the base of the output.
        The result is a subset of this dictionary.
    keys : iterable(immutable)
        Effectively a list of the keys to extract from source
        (though any iterable should do).
        The keys do not need to be in source, and only those also
        in source will be in the result.
        
    Result
    ------
    dict
        The dict of key value pairs based on the intersection of 
        the source's keys and keys.
        
    Examples
    --------
    >>> from pyfac.utils import *  # doctest: +SKIP
    >>> options = {'flag':True, 'int':5, 'float':3.5,'array':np.arange(5)}
    >>> maybe_pops(options,['flag'])
    {'flag': True}
    >>> maybe_pops(options,['int','float'])
    {'int': 5, 'float': 3.5}
    >>> maybe_pops(options,['flag','array'])
    {'array': array([0, 1, 2, 3, 4])}
    >>> options
    {}
    """

    return {key: source.pop(key) for key in keys if key in source}


def maybe_keys(source, keys):
    """ Create a new dict, by poping the intersection of source's keys and keys.
    
    Create a new dictionary. Each pair of key and value comes from source.
    Inclusion in the result is based on intersection of source's keys and keys.
    This function has the side-effect that the key-value pairs in the result
    are removed from source (poped).
    
    Parameters
    ----------
    source : dict
        A dictionary forming the base of the output.
        The result is a subset of this dictionary.
        There (may) be side-effect on this, of the form that some
        key-value pairs (may) be removed.
    keys : iterable(immutable)
        Effectively a list of the keys to extract from source
        (though any iterable should do).
        The keys do not need to be in source, and only those also
        in source will be in the result.
        
    Result
    ------
    dict
        The dict of key value pairs based on the intersection of 
        the source's keys and keys.
        
    Examples
    --------
    >>> from pyfac.utils import *  # doctest: +SKIP
    >>> options = {'flag':True, 'int':5, 'float':3.5,'array':np.arange(5)}
    >>> maybe_keys(options,['flag'])
    {'flag': True}
    >>> maybe_keys(options,['int','float'])
    {'int': 5, 'float': 3.5}
    >>> maybe_keys(options,['flag','array'])
    {'flag': True, 'array': array([0, 1, 2, 3, 4])}
    >>> options
    {'flag': True, 'int': 5, 'float': 3.5, 'array': array([0, 1, 2, 3, 4])}
    """
    return {key: source[key] for key in keys if key in source}


def call(function, dict_, *args, use_pop=None, **kwargs):
    """ Call a function with arguments mixed from different sources.
    
    Call a function, by mixing the arguments for the function.
    Specifically it extract extra parameters from `dict_` 
    to use as default keyword arguments in case they have not
    been specified otherwise.
    If the function called accepts `**kwargs`, then all `dict_`
    arguments not otherwise relevant will also be added,
    so they can be passed further down possible call chains.
    
    Parameters
    ----------
    function : a -> b
        A function to be called with the composed arguments.
        Note that the argument names for this function is used
        to determine which keyword arguments to extract from `dict_`
    `dict_` : dict
        A directory of keyword argument pairs that might be 
        keyword arguments for the function.
        There is no requirement for any overlap with argument names
        of function, but only the overlap is used here.
    *args
        The positional arguments to be passed to function.
    use_pop : boolean, optional
        A flag for whether the arguments in `dict_` used in the call
        should be removed from dict.
        Defaults to False (None, which evaluates to False).
    **kwargs
        The keyword arguments to be passed to the function.
        These keyword arguments takes priority over those in `_dict`.
        
    Returns
    -------
    ?
        The result of the function call.
        
    Examples
    --------
    Base use:
    >>> from pyfac.utils import *  # doctest: +SKIP
    >>> options = {'password':'MyPassword','IV':16*'\x0F','target_index':1}
    >>> call(safe_user.box_content,options,'bla') # doctest: +ELLIPSIS
    b'...'
    """
    names = list(function.__code__.co_varnames[:function.__code__.co_argcount])
    residual = [name for name in names[len(args):] if name not in kwargs]
    maybe_get = maybe_pops if use_pop else maybe_keys
    options = maybe_get(dict_, residual)
    if function.__code__.co_flags & 0x08: #this flag is true if function as a **kwargs
        # consider the optimization of making a set(names) for faster lookups below
        options.update({k:v for k,v in dict_.items() if k not in names})
        # consider whether use_pop should be inheireted
        # if use_pop is not None:
            # options.update({'use_pop':use_pop})
    return function(*args, **kwargs, **options)
    
    
# def _pack_nd(*scalarss):
    # """ Pack n arrays of scalars into an array of n dimensional vectors.
    # """
    # if len(scalarss) < 1:
        # return None
    # elif len(scalarss) == 1:
        # return np.expand_dims(scalarss[0],1)
    # dimensions = len(scalarss)
    # result = np.zeros((len(scalarss[0]),dimensions))
    # for i, scalars in enumerate(scalarss):
        # result[:,i] = scalars
    # return results

    
def pack_3d(xs, ys, zs):
    """ Pack 3 array-likes into an array of 3d vectors.
    
    Parameters
    ----------
    xs : array-like
        The first (xs) components of the final vector.
        Must be equivalent to an array of scalars,
        generators are not accepted (though ranges are).
        Must have a reasoable response to len(xs),
        which must be the same as len(ys) and len(zs).
    ys : array-like
        The second (ys) components of the final vector.
        Must be equivalent to an array of scalars,
        generators are not accepted.
        Must have a reasoable response to len(ys),
        which must be the same as len(xs) and len(zs).
    zs : array-like
        The third (zs) components of the final vector.
        Must be equivalent to an array of scalars,
        generators are not accepted.
        Must have a reasoable response to len(zs),
        which must be the same as len(ys) and len(xs).
        
    Returns
    -------
    ndarray
        An array of vectors each with 3 dimensions,
        and the same length as xs.
        
    Examples
    --------
    >>> from pyfac.utils import *  # doctest: +SKIP
    >>> pack_3d(range(5),np.arange(5)*2-3,[np.cos(xs*np.pi) for xs in range(5)])
    array([[ 0., -3.,  1.],
           [ 1., -1., -1.],
           [ 2.,  1.,  1.],
           [ 3.,  3., -1.],
           [ 4.,  5.,  1.]])
    """
    # return _pack_nd(xs,ys,zs)
    result = np.zeros((len(xs), 3))
    # if (len(xs) != len(ys) or len(xs) != len(zs)):
        # raise ValueError('lengths not equal, with: '
                         # + '!='.join(map(lambda i: str(len(i)),[xs,ys,zs])))
    result[:, 0] = xs
    result[:, 1] = ys
    result[:, 2] = zs
    return result
    
    
# def _as_nd(scalars, n=3):
    # """ Copies an array of scalars so it becomes an array of vectors.
    # """
    # return pack_3d(*[scalars]*n)


def as_3d(scalars):
    """ As pack_3d, but with the same scalars in on all components.
    
    Parameters
    ----------
    scalars : array-like
        An array-like of scalars.
        
    Returns
    -------
    ndarray
        An array of 3 dimensional vectors with the same length as xs.
        
    Examples
    --------
    >>> from pyfac.utils import *  # doctest: +SKIP
    >>> vec = np.arange(15).reshape(5,3)
    >>> as_3d(range(5)) * vec
    array([[ 0.,  0.,  0.],
           [ 3.,  4.,  5.],
           [12., 14., 16.],
           [27., 30., 33.],
           [48., 52., 56.]])
    """
    # return _as_nd(xs)
    # """ As pack_3d, but with the same vector in on all components.
    # Use this form array operations where one part has a scalars (this),
    # and another 3d vectors.
    # """
    return pack_3d(scalars, scalars, scalars)

    
# # currently unused, alternative to NEC_to_VSC,
# # in consideration as replacement
# class _NEC_to_VSC2:
    
    # def __init__(self, velocities=None,angles=None,sines_cosines = None):
        # # self.ready = False
        # if sines_cosines is not None:
            # self.sines, self.cosines = sines_cosines
            # # self.ready = True
        # elif angles is not None:
            # self.from_angles(angles)
        # elif velocities is not None
            # self.from_velocities(velocities)
        # else:
            # raise ValueError("At least one parameter must be not None")
        
    # def from_angles(self,angles):
        # self.sines = np.sin(angles)
        # self.cosines = np.cos(angles)
        # # self.ready = True
        
    # def from_velocities(self, velocities):
        # self.from_angles(-np.arctan2(velocities[:, 0] - velocities[:, 1],
                                     # velocities[:, 0] + velocities[:, 1]))
                                     
    # def __call__(self,vectors):
        # return pack_3d(self.cosines*vectors[:, 0] + self.sines*vectors[:, 1],
                       # - self.sines*vectors[:, 0] + self.cosines*vectors[:, 1],
                       # vectors[:, 2])
                       
    # def inverse(self):
        # self.sines = -self.sines
        # # return pack_3d(self.cosines*vectors[:, 0] - self.sines*vectors[:, 1],
                       # # self.sines*vectors[:, 0] + self.cosines*vectors[:, 1],
                       # # vectors[:, 2])
    

def map_3d(function,vectors):
    """ map_3d maps a function for 1d data to each dimension in 3d data.
    
    Parameters
    ----------
    function : ndarray -> ndarray
        A function that takes a scalar array
        and maps it to another scalar array.
    vectors : ndarray
        An array of 3 dimensional vectors.
        
    Returns
    -------
    ndarray
        An array of 3 dimensional vectors.
        
    Examples
    --------
    >>> from pyfac.utils import *  # doctest: +SKIP
    >>> vec = np.arange(15).reshape(5,3)
    >>> map_3d(lambda xs: [x > 2 and x%4 != 0 for x in xs],vec)
    array([[0., 0., 0.],
           [1., 0., 1.],
           [1., 1., 0.],
           [1., 1., 1.],
           [0., 1., 1.]])
    """
    return pack_3d(*[function(vectors[:,i]) for i in range(3)])

def NEC_to_VSC(velocities):
    """ Construct a function to transform from the NEC frame to the VSC frame.
    
    This function generates another function,
    which is a transform of vectors from the NEC frame to the VSC frame.
    This specific transformation may be different for each data point.
    
    Parameters
    ----------
    velocities : ndarray
        The velocities describe thw forward direction in the VSC frame,
        and they can be different for each data point.
    
    Returns
    -------
    function
        A function that takes a list of vectors
        of the same length as velocities and
        transform them from the NEC frame to the VSC frame.
        
    Examples
    --------
    >>> from pyfac.utils import *  # doctest: +SKIP
    >>> v = np.arange(15).reshape(5,3)
    >>> trans = NEC_to_VSC(v)
    >>> trans(v) # doctest: +ELLIPSIS
    array([[ 0.7...,  0.7...,  2. ...],
           [ 3.5...,  3.5...,  5. ...],
           [ 6.5...,  6.5...,  8. ...],
           [ 9.5...,  9.5..., 11. ...],
           [12.5..., 12.5..., 14. ...]])
    >>> trans((v[::-1]-2.5) * (v + 3)) # doctest: +ELLIPSIS
    array([[ 49.8...,   9.5...,  57.5...],
           [ 46.0...,  46.4...,  68. ...],
           [ 34.8...,  42.4...,  60.5...],
           [  7.0...,  19.1...,  35. ...],
           [-38.4..., -22.4...,  -8.5...]])
    """
    # """ Build a transform going from the NEC frame to the VEC frame.
    # This is done based on the velocities at each point, so the
    # transformation will act differently on each point.
    # Returns the actual transformation function,
    # which can then be applied to compariable series.
    # """
    angles = -np.arctan2(velocities[:, 0] - velocities[:, 1],
                         velocities[:, 0] + velocities[:, 1])
    sines = np.sin(angles)
    cosines = np.cos(angles)

    def transform(vectors):
        """ Transforms from the NEC frame to the VSC frame
        
        This function takes an array of vectors,
        which is in the NEC frame and returns them transformed
        into the VSC frame.
        The transform will may be different for each vector,
        and is fixed when this function is constructed.
        
        Parameters
        ----------
        vectors : ndarray
            An array of 3d points or vectors in the NEC frame.
            
        Returns
        -------
        ndarray
            An array of the vectors in the VSC frame.
            
        Note
        ----
        See NEC_to_VSC for examples of use.
        """
        # """ Transforms from the NEC frame to the VEC frame.
        # Note that the frames have alligned origos,
        # so vectors and points transform the same way.
        # """
        return pack_3d(cosines*vectors[:, 0] + sines*vectors[:, 1],
                       - sines*vectors[:, 0] + cosines*vectors[:, 1],
                       vectors[:, 2])  # Does the 3rd dimension flip?
    return transform


def delta(vectors):
    """ Computes the finite difference on an arralike.

    Parameters
    ----------
    vectors : ndarray
        An array of vectors or scalars.
        
    Returns
    -------
    ndarray
        An array of vectors or scalars of shape (n-1,...),
        when vectors input have shape (n,...).
    
    Examples
    --------
    >>> from pyfac.utils import *  # doctest: +SKIP
    >>> delta(np.arange(5))
    array([1, 1, 1, 1])
    >>> delta(np.array([4, 3, 7, 4, 2, 92]))
    array([-1,  4, -3, -2, 90])
    >>> delta(np.array([[1, 0, 0], [0, 1, 3], [7, -13, 4]]))
    array([[ -1,   1,   3],
           [  7, -14,   1]])
    """
    return vectors[1:] - vectors[:-1]


def means(vectors):
    """ Computes the means between this and the next point.

    Parameters
    ----------
    vectors : ndarray
        An array of vectors or scalars.
        
    Returns
    -------
    ndarray
        An array of vectors or scalars of shape (n-1,...),
        when vectors input have shape (n,...).
    
    Examples
    --------
    >>> from pyfac.utils import *  # doctest: +SKIP
    >>> means(np.array(list(range(5))))
    array([0.5, 1.5, 2.5, 3.5])
    >>> means(np.array([4, 3, 7, 4, 2, 92]))
    array([ 3.5,  5. ,  5.5,  3. , 47. ])
    >>> means(np.array([[1, 0, 0], [0, 1, 3], [7, -13, 4]]))
    array([[ 0.5,  0.5,  1.5],
           [ 3.5, -6. ,  3.5]])
    """
    return (vectors[1:] + vectors[:-1]) * 0.5


def spherical_delta(positions):
    """ Computes the change of sherical positions as a NEC vector.
    
    Parameters
    ----------
    positions : ndarray
        An array of positions (3d vectors) described in spherical coordinates.
        
    Returns
    -------
    ndarray : ndarray
        An array of vectors in the NEC frame.
        
    Examples
    --------
    >>> from pyfac.utils import *  # doctest: +SKIP
    >>> vecs = np.arange(15).reshape(5,3)
    >>> spherical_delta(vecs)
    array([[ 0.18317585,  0.18311308, -3.        ],
           [ 0.34018372,  0.33913504, -3.        ],
           [ 0.49719158,  0.49293804, -3.        ],
           [ 0.65419945,  0.64324482, -3.        ]])
    """
    theta, Lambda, r = (positions[:, i] for i in range(3))
    r_means = means(r)
    return pack_3d(r_means * np.sin(TO_RADIANS * delta(theta)),
                   r_means * np.sin(TO_RADIANS * delta(Lambda))
                           * np.cos(TO_RADIANS * means(theta)),
                   -delta(r))

  
                   
def curl(delta_x, delta_field, target_index=2):
         # delta_left=None, delta_right=None, target_index=2):
    """ Compute a single component of curl for an array of vectors.
    
    This function is used to calculate a finite difference apporximation
    of curl. Its starting point is similar to a finite difference 
    approximation of a partial derivative, where you use both the 
    recoded change in the function and what you differentiate with respect to.
    In practice this calculate an array of in independent curl operations,
    as given by a single step.
    
    Parameters
    ----------
    delta_x : ndarray
        An array of the finite difference in the considered step.
    delta_field : ndarray
        An array of the finite difference of the field under
        the relevant considered step.
    target_index : int, optional
        The index of the dimension of the curl that should be given as output.
        
    Returns
    -------
    ndarray
        An array of the curl-like quantity,
        but only target_index dimension of it.
        
    Examples
    --------
    >>> from pyfac.utils import *  # doctest: +SKIP
    >>> vecs = delta(np.arange(15).reshape(5,3))
    >>> curl(vecs,vecs)
    array([0., 0., 0., 0.])
    >>> curl(pack_3d(np.ones((4,)),-1.5 * np.ones((4,)),np.zeros((4,))),vecs)
    array([5., 5., 5., 5.])
    """
    index_x = (target_index+1) % 3
    index_y = (target_index+2) % 3
    # if delta_left is None:
        # if left is None:
            # raise ValueError("Either left or delta_left must be specified.")
        # delta_left = delta(left)
    # if delta_right is None:
        # if right is None:
            # raise ValueError("Either right or delta_right must be specified.")
        # delta_right = delta(right)
    return (delta_field[:, index_y] / delta_x[:, index_x]
            - delta_field[:, index_x] / delta_x[:, index_y])


def inclination(vectors):
    """ Calculates the inclination of vectors.
    
    Calculate the inclination for each individual vector in vectors.
    This means this is equivalent to the tilt of the vectors toward
    the third dimention.
    
    Parameters
    ----------
    vectors : ndarray
        An array of 3d vectors.
        
    Returns
    -------
    ndarray
        An array of scalars, where each is the inclination
        of their respective vector in vectors.
        
    Examples
    --------
    >>> from pyfac.utils import *  # doctest: +SKIP
    >>> inclination(np.arange(15).reshape(5,3))
    array([1.10714872, 0.78539816, 0.71469295, 0.68539502, 0.66942998])
    >>> np.sin(_)
    array([0.89442719, 0.70710678, 0.65538554, 0.63297887, 0.62053909])
    """
    length_first_second_dims = (vectors[:, 0]**2 + vectors[:, 1]**2)**0.5
    return np.arctan2(vectors[:, 2], length_first_second_dims)


def request_data(start=date.datetime(2016, 1, 1),
                 end=date.datetime(2016, 1, 2),
                 user_file='safe_xml.txt',
                 credentials = None,
                 password = None,
                 target_file='tempdata.cdf',
                 filters=[{'parameter': 'Latitude',
                           'minimum': -90.,
                           'maximum': 90.}],
                 target_url='https://staging.viresdisc.vires.services/openows',
                 collection='SW_OPER_MAGA_LR_1B',
                 product_options={'auxiliaries': ['QDLat', 'QDLon']},
                 sampling_step='PT1S',
                 models=['MCO_SHA_2C', 'MLI_SHA_2C',
                         'MMA_SHA_2C-Primary', 'MMA_SHA_2C-Secondary'],
                 measurements=['F', 'B_NEC']):
    """ Request data from a vires server.
    
    This function sets up a request to a vires server.
    It is a wrapper on viresclient functionality,
    where default arguments have been provided.
    
    The data will be saved to as a cdf file.
    
    Parameters
    ----------
    start : datatime, optional
        The inclusive starting time of the requested data.
        Defaults to 1st of Janurary 2016
    end : datetime, optional
        The exclusive end time of the requested data.
        Defaults to 2nd of Janurary 2016
    user_file : str, optional
        Filepath to the safe_user encrypted file with login credentials.
        If set to None you will be prompted for login credentials in runtime.
        Defaults to 'safe_xml.txt'
    password : str, optional
        The password needed to decrypt user_file.
        If not provided/set to None you will be prompted for it in runtime.
    target_file : str, optional
        Filepath for the output cdf file. Must include name and extension.
        Defaults to 'tempdata.cdf'
    filters : list(dict), optional
        A list of dictionaries, where each dictionary is the options
        for a filter to be applied to the data.
    target_url : str, optional
        The full url for the server to request the data at.
        Defaults to 'https://staging.viresdisc.vires.services/openows'
    collection : str, optional
        See viresclient set_collection. Defaults to 'SW_OPER_MAGA_LR_1B'.
    product_options : dict, optional
        Extra options to viresclient set_products.
        Defaults to {'auxiliaries': ['QDLat', 'QDLon']}
    sampling_step : str, optional
        Describes the sampling frequency, 
        since vires may otherwise downsample the data.
        The default is 1Hz by 'PT1S'.
    models : list(str), optional
        The models calculated at the data points.
        See viresclient set_products models for details.
        Defaults to [
        'MCO_SHA_2C', 'MLI_SHA_2C',
        'MMA_SHA_2C-Primary', 'MMA_SHA_2C-Secondary']
    measurements : list(str), optional
        Measured quantities to be included for each data point.
        Defaults to ['F', 'B_NEC']
    
    Examples
    --------
    >>> from pyfac.utils import *  # doctest: +SKIP
    >>> request_data() # doctest: +SKIP
    >>> request_data(date.datetime(2017, 7, 5),
    ...              date.datetime(2017, 7, 6))# doctest: +SKIP
                     
    Note
    ----
    These examples are skiped by doctest due to security reasons,
    and unintentional side-effects.
    """
    if credentials is None:
        if user_file is not None:
            credentials = safe_user.fetch_login('safe_xml.txt')
        else:
            (username, password) = safe_user.prompt_login_data()
            credentials = {'username':username,'password':password}
    request = SwarmRequest(url=target_url, **credentials)
                           # **safe_user.fetch_login('safe_xml.txt'))
    request.set_collection(collection)
    request.set_products(measurements=measurements, models=models,
                         sampling_step=sampling_step, **product_options)
    for filter in filters:
        request.set_range_filter(**filter)
    data = request.get_between(start_time=start, end_time=end)
    data.to_file(target_file)


def read_cdf(file, **name_pairings):
    """ Read data from a cdf file.
    
    Read a cdf file and construct a dictionary based on
    name_parings. This may open and read a file as a side-effect.
    
    Parameters
    ----------
    file : str or cdfread
        The file to read the data from. 
        Accepts both a string filepath to the file and an opened cdfread file.
    **name_pairings
        Pairs of names used to link a name in the output with its 
        corrisponding name in the cdf file.
        
    Returns
    -------
    dict
        The keys are the keywords used in name_parings,
        while the values are the variables loaded from
        the cdf file with the name of the corrisponding argument.
        
    Examples
    --------
    We can directly ask for variables:
    
    >>> from pyfac.utils import *  # doctest: +SKIP
    >>> read_cdf('tempdata.cdf') # doctest: +SKIP
    {}
    >>> read_cdf('tempdata.cdf', time='Timestamp') # doctest: +SKIP,+ELLIPSIS
    {'time': array([...])}
    
    We can also make use of predefined options:
    
    >>> options = {'time': 'Timestamp', 'B_base': 'B_NEC'}
    >>> read_cdf('tempdata.cdf', **options) # doctest: +SKIP,+ELLIPSIS
    {'time': array([...]), 'B_base': array([[...]])}
    
    Note
    ----
    These examples are skiped by doctest due to dependency 
    on the existance of a 'tempdata.cdf' file.
    """
    if not isinstance(file, cdflib.cdfread.CDF):
        file = cdflib.CDF(file)
    return {name: file.varget(cdf_name)
            for name, cdf_name in name_pairings.items()}


if __name__ == '__main__':
    # Test docstrings
    import doctest
    doctest.testmod()



""" The fac module is used for calculated field alligned currents.

This module contains functions used to calculated field alligned currents,
and some of the components used in the calculations.
This module is dependent on the swarm_util module, which contains
functions for utilities and some of the more general components needed.

This module currently only allows caluculation of the single satelite
version of the FAC product, and it uses data available from viresclient's
python client or another data source with similar format.
The major difference between this and the official ESA product,
is that the input from models are their effects at each measurement point,
since this is the format delivered by viresclient.

This means that there is a small algorithmic difference compared
to what is used in the official ESA products,
though experimental comparison have found the difference 
to be less than  1.2*10^(-4) micro A/m^2, and the average
difference to be less than 1% of this.

Running this module as a script will (outside of running some tests)
perform the default experiment.

Overview
--------
This section lists each of the functions in this module,
and give a short description of them.
You can use this as an easier way to look up what you might need.

split_models
    Takes measurements and up to several models,
    and returns the combined model and the residual on the measurements.
radial_current
    Calculate the radial currents for time-like series.
single_sat_fac
    Calculate the field alligned currents
    for a single satelinte on time-like series.
single_sat_fac_full
    Like single_sat_fac, but it handles input from a dictionary,
    and can figure out which models are pressent itself.
fetch_data
    Fetch the data needed, and can figure out itself whether
    it needs to download it first.
fac_from_file
    Do the entire single satelite field alligned currents calculations,
    but takes a possible file as input the same way as fetch_data.

Examples
--------
>>> from pyfac.fac import fac_from_file  # doctest: +SKIP
>>> import datetime as date
>>> results, input_data = fac_from_file(
...     start=date.datetime(2017, 4, 11),
...     end=date.datetime(2017, 4, 12),
...     force_download=True)  # doctest: +SKIP
>>> lat, fac = results[1][:, 0], results[3]  # doctest: +SKIP
"""
# """ A slim version of the FAC level 2 processing,
# but when using viresclient data instead of official ESA products.
# """

__version__ = '0.5.0'
__author__ = 'Ask Neve Gamby'

import os
import numpy as np
import matplotlib.pyplot as plt
import scipy.interpolate as sii

from . import safety as safe_user
from . import utils as sw

_INKL_LIMIT = 30.

base_pairs = {
    'time': 'Timestamp',
    'Theta': 'Latitude', 
    'phi': 'Longitude', 
    'r': 'Radius'
    }

_name_pairings = {
    **base_pairs,
    'B_base': 'B_NEC',
    'B_core': 'B_NEC_MCO_SHA_2C',
    'B_lithosphere': 'B_NEC_MLI_SHA_2C',
    'B_magnetosphere_primary': 'B_NEC_MMA_SHA_2C-Primary',
    'B_magnetosphere_secondary': 'B_NEC_MMA_SHA_2C-Secondary',
    }


def split_models(baseline, *models):
    """ Seperate the effect of models from the baseline.
    
    Seperate the effect of model and risidual on the baseline,
    base on precalculated effects of each model.
    It is assumed that the models combine linearly with each
    other and a residual (meaning modelA + modelB + residual = baseline,
    but not modelA * modelB + residual = baseline or other such variations).
    
    Parameters
    ----------
    baseline : ndarray
        An array of the base measurements.
    *models : ndarray
        Arrays of the effects of each model to be removed.
        all models must have the same length as baseline.
    
    Returns
    -------
    tuple of ndarray
        A tuple, with the first component being an ndarray of the residual
        and the second component being an ndarray of the combined model.
        The result statisfies that the sum of the components
        is equal to the baseline.
        
    Examples
    --------
    >>> from pyfac.fac import *  # doctest: +SKIP
    >>> modelA = np.arange(15).reshape(5,3)**2
    >>> modelB = sw.as_3d(np.arange(7,2,-1))
    >>> data = (np.random.randn(5,3)*0.1 + modelA + modelB 
    ...         + sw.pack_3d(np.sin(np.arange(5)*0.5),
    ...                      np.cos(np.arange(5)*0.5),
    ...                      np.tan(np.arange(5)*0.5)))
    >>> split_models(data,modelA,modelB) # doctest: +ELLIPSIS
    (array([[...,...,...],
           [...,...,...],
           [...,...,...],
           [...,...,...],
           [...,...,...]]), array([[  7.,   8.,  11.],
           [ 15.,  22.,  31.],
           [ 41.,  54.,  69.],
           [ 85., 104., 125.],
           [147., 172., 199.]]))
    """
    model = sum(models)
    return (baseline - model,
            model)


def radial_current(delta_B, velocity, delta_time, MU_0=sw.MU_0):
    """ Calculate the radial curents (IRC) over a time series.
    
    Parameters
    ----------
    delta_B : ndarray
        The change in the B (magnetic) field in a time series in
        an ndarray of vectors.
        Standard use requires this to be in the VSC frame.
    velocity : ndarray
        An ndarray of vectors of the spartial velocities
        at each point in the timeseries.
    delta_time : ndarray
        An ndarray of scalars of the time steps between 
        each point in the timeseries of delta_B and velocity.
    MU_0 : float, optional
        The constant of permubility of vacum, in the given units.
        Defaults to swarm_util.MU_0, which assumes velocity 
        and delta_time are in base SI units, and delta_B in nT.
        
    Returns
    -------
    ndarray
        An ndarray of vectors, each of which describes
        the radial current at the respective point in the timseries.
        
    Examples
    --------
    >>> from pyfac.fac import *  # doctest: +SKIP
    >>> vecs = sw.delta(np.arange(15).reshape(5,3))
    >>> radial_current(vecs,vecs,np.ones(len(vecs)))
    array([-0., -0., -0., -0.])
    >>> radial_current(vecs, sw.pack_3d(np.ones((4,)),-1.5 * np.ones((4,)),np.zeros((4,))),np.ones(len(vecs)))
    array([-1989.43678865, -1989.43678865, -1989.43678865, -1989.43678865])
    """
    change = sw.curl(delta_x=velocity, delta_field=delta_B)
    return ((-.001 / (2*MU_0)) * change / delta_time)


def single_sat_fac(time, positions, B_res, B_model):
    """ Calculate the Field Aligned Currents on a single satelites path.
    
    This function runs through the same logic as needed for the single
    satelite version of the official FAC product.
    This function is not designed to deal with data flaged in some way as bad.
    There are several extra outputs usefull for debuging and comparisons.
    
    Parameters
    ----------
    time : ndarray
        An array of scalars of time stamps in seconds.
    poisitions : ndarray
        An array of vectors (theta, phi,r) of the position of the 
        satelite in spherical coordinates.
    B_res : ndarray
        An array of vectors of the residual of the magnetic field 
        measurements after the model have been substracted.
        The vectors are assumed to be in the NEC coordinate system.
    B_model : ndarray
        An array of vectors of the model of the magnetic field
        at the point of measurements.
        The vectors are assumed to be in the NEC coordinate system.
        
    Returns
    -------
    The return is a tuple consiting of the following:
    time : ndarray
        An ndarray of  scalars describing the time each point
        in the calculations corrisond to.
        Note that it is off compared to the parameter time.
    position : ndarray
        An ndarray of spherical coordinates describing the postion
        of each point in the time series where the calculations
        are suppose to apply at. The coordinates are arranged as
        theta, phi, r.
        Note tht this is off compared to the parameter positions.
    irc : ndarray
        An ndarray of vectors of the radial currents (the radial components,
        of currents) found when solving one of Maxwells equations
        on the magnetic field.
        Note that this is given in the VSC (spacecraft velocity) frame.
    fac : ndarray
        An ndarray of vectors of the field alligned currents.
        Note that this is given in the VSC (spacecraft velocity) frame.
    V : ndarray
        An ndarray of vectors of the velocity in the NEC frame.
    V_VSC : ndarray
        An ndarray of vectors of the velocity in the VSC frame.
    pos_ltl : ndarray
        An ndarray of scalars of positions where local time is acounted for.
        Each position reference the same position as positions output.
        Note tht this is off compared to the parameter positions.
    inclination :
        An ndarray of scalars of the inclination of the aproximation 
        of the magnetic field, where the calculations
        are supposed to be applied at.
        
    Examples
    --------
    >>> from pyfac.fac import *  # doctest: +SKIP
    >>> time = np.arange(0,10,0.1)
    >>> positions = sw.pack_3d(np.sin(0.01 * time)*90,(0.01 * time % np.pi)/sw.TO_RADIANS ,np.ones(len(time)))
    >>> B_res = np.random.randn(len(time),3)* sw.as_3d(10 *positions[:,1])
    >>> B_model = sw.as_3d(positions[:,2])*sw.pack_3d(np.sin(positions[:,0]),np.zeros(len(time)),2*np.cos(positions[:,0]))
    >>> single_sat_fac(time,positions,B_res,B_model)  # doctest: +ELLIPSIS
    (array([...]), array([[...]]), array([...]), array([...]), array([[...]]), array([[...]]), array([[...]]), array([...]))
    """
    dt = sw.delta(time)
    pos_ltl = sw.pack_3d(positions[:,0],
                         ((positions[:,1] + 180 + time/86400 * 360)%360)-180,
                         positions[:,2])
    V = sw.spherical_delta(pos_ltl)/sw.as_3d(dt)
    transform = sw.NEC_to_VSC(V)
    V_VSC = transform(V)  # V is in VSC frame now
    B1 = transform(B_res[:-1])
    B2 = transform(B_res[1:])
    #transform(sw.delta(B_res)
    irc = radial_current(B2-B1, V_VSC, dt)
    target_time = sw.means(time)
    B_interpolated = sw.map_3d(lambda data: 
        sii.splev(target_time,sii.splrep(time,data)),
        B_model)
    inclination = sw.inclination(B_interpolated) #we approximate local field by interpolation
    fac = -irc / np.sin(inclination)
    fac[abs(inclination) < _INKL_LIMIT * sw.TO_RADIANS] = np.nan
    is_good = dt <= 1
    return (sw.means(time)[is_good],
            sw.means(positions)[is_good],
            irc[is_good],
            fac[is_good],
            V[is_good],
            V_VSC[is_good],
            sw.means(pos_ltl)[is_good],
            inclination[is_good])


def single_sat_fac_full(input_data):
    """ Calculate the Field Aligned Currents on a single satelites path.
    
    This is mainly designed as a wrapper for single_sat_fac,
    which can extract the relevant parameters from a dictionary
    
    This function runs through the same logic as needed for the single
    satelite version of the official FAC product.
    This function is not designed to deal with data flaged in some way as bad.
    There are several extra outputs usefull for debuging and comparisons.
    
    Parameters
    ----------
    There is only one parameter: input_data, which is a dictionary.
    This dictionary functions effectively as a superset of
    the relevant parameters.  The relevant parameters are given below:
    time : ndarray
        An array of scalars of time stamps in seconds.
    Theta : ndarray
        An array of scalars representing the polar angle component
        of the satelite in spherical coordinates.
    phi : ndarray
        An array of scalars representing the azimuth angle component
        of the satelite in spherical coordinates.
    r : ndarray
        An array of scalars representing the radial angle component
        of the satelite in spherical coordinates.
    B_base : ndarray
        An array of vectors of the magnetic field measurements.
        The vectors are assumed to be in the NEC coordinate system.
    B_core : ndarray, optional
        An array of vectors of the magnetic field component
        derived from a model of Earth's core.
    B_lithosphere : ndarray, optional
        An array of vectors of the magnetic field component
        derived from a model of the Earthø's lithosphere (crust).
    B_magnetosphere_primary : ndarray, optional
        An array of vectors of the magnetic field component
        derived from a model of the primary magnetosphere field.
    B_magnetosphere_secondary : ndarray, optional
        An array of vectors of the magnetic field component
        derived from a model of the induced field from the magnetosphere.
        
    Returns
    -------
    The return is a tuple consiting of the following:
    time : ndarray
        An ndarray of  scalars describing the time each point
        in the calculations corrisond to.
        Note that it is off compared to the parameter time.
    position : ndarray
        An ndarray of spherical coordinates describing the postion
        of each point in the time series where the calculations
        are suppose to apply at. The coordinates are arranged as
        theta, phi, r.
        Note tht this is off compared to the parameter positions.
    irc : ndarray
        An ndarray of vectors of the radial currents (the radial components,
        of currents) found when solving one of Maxwells equations
        on the magnetic field.
        Note that this is given in the VSC (spacecraft velocity) frame.
    fac : ndarray
        An ndarray of vectors of the field alligned currents.
        Note that this is given in the VSC (spacecraft velocity) frame.
    V : ndarray
        An ndarray of vectors of the velocity in the NEC frame.
    V_VSC : ndarray
        An ndarray of vectors of the velocity in the VSC frame.
    pos_ltl : ndarray
        An ndarray of scalars of positions where local time is acounted for.
        Each position reference the same position as positions output.
        Note tht this is off compared to the parameter positions.
    inclination :
        An ndarray of scalars of the inclination of the aproximation 
        of the magnetic field, where the calculations
        are supposed to be applied at.
        
    Examples
    --------
    >>> from pyfac.fac import *  # doctest: +SKIP
    >>> time = np.arange(0,10,0.1)
    >>> positions = sw.pack_3d(np.sin(0.01 * time)*90,(0.01 * time % np.pi)/sw.TO_RADIANS ,np.ones(len(time)))
    >>> B_res = np.random.randn(len(time),3)* sw.as_3d(10 *positions[:,1])
    >>> B_core = sw.as_3d(positions[:,2])*sw.pack_3d(np.sin(positions[:,0]),np.zeros(len(time)),2*np.cos(positions[:,0]))
    >>> single_sat_fac_full({'time':time,'Theta':positions[:,0],'phi':positions[:,1],'r':positions[:,2],'B_base':B_res+B_core,'B_core':B_core})  # doctest: +ELLIPSIS
    (array([...]), array([[...]]), array([...]), array([...]), array([[...]]), array([[...]]), array([[...]]), array([...]))
    """
    return single_sat_fac(
        input_data['time'],
        sw.pack_3d(input_data['Theta'], input_data['phi'], input_data['r']),
        *split_models(input_data['B_base'], *[input_data[key] for key in [
            'B_core', 'B_lithosphere',
            'B_magnetosphere_primary',
            'B_magnetosphere_secondary'] if key in input_data]))


def fetch_data(force_download=False, temp_file='tempdata.cdf',**options):
    """ Fetch cdf form data from a file or by downloading it.
    
    This function is made to hide the complexity of getting
    data from a cdf file that you might need to download first.
    The idea is a dowloaded file will be saved as some temporary file
    temp_file and then opened.
    If the file have already been downloaded then it will be assumed
    that the file at temp_file is the one wanted, unless
    force_download is set to true to indicate that this cache
    should be flushed and overwritten by a new download.
    
    Parameters
    ----------
    force_download : boolean, optional
        A flag on whether to force downloading of the data and
        thereby overwrite the temp_file if it already exists.
        If set to False the program will only download the data
        if it cannot find the temp_file.
        Default is False.
    temp_file : string
        A string of the file path for where the downloaded file should be.
        If the file already exists then this may take priority
        depending on the force_download flag.
        
    Returns
    -------
    dict
        A dictionary map of _name_pairings, where the values
        have been replaced with the data entry of
        the previous value.
    
    Examples
    --------
    >>> from pyfac.fac import *  # doctest: +SKIP
    >>> fetch_data()  # doctest: +ELLIPSIS, +SKIP
    {...}
    """
    name_pairings = options.pop('name_pairings',_name_pairings)
    if temp_file is None or not os.path.isfile(temp_file) or force_download:
        sw.call(sw.request_data,options,target_file=temp_file)
    return sw.read_cdf(temp_file, **name_pairings)


def fac_from_file(**options):
    """ Compute FAC for a single satelite from data in an outside source.
    
    This function is inteded as the topmost function of this module.
    It is therefore suggested to just strait up use this function,
    when one wants to use this as a module instead of as a script.
    
    The function will read a cdf-file or download it,
    if such a file is not already cached.
    It will then calculate FAC (field aligned currents),
    and other related measures, and return those results.
    
    Parameters
    ----------
    **options
        options to be passed to other lower level function(s),
        mainly fetch_data.
        
    Returns
    -------
    The return is a tuple consiting of the following:
    time : ndarray
        An ndarray of  scalars describing the time each point
        in the calculations corrisond to.
        Note that it is off compared to the parameter time.
    position : ndarray
        An ndarray of spherical coordinates describing the postion
        of each point in the time series where the calculations
        are suppose to apply at. The coordinates are arranged as
        theta, phi, r.
        Note tht this is off compared to the parameter positions.
    irc : ndarray
        An ndarray of vectors of the radial currents (the radial components,
        of currents) found when solving one of Maxwells equations
        on the magnetic field.
        Note that this is given in the VSC (spacecraft velocity) frame.
    fac : ndarray
        An ndarray of vectors of the field alligned currents.
        Note that this is given in the VSC (spacecraft velocity) frame.
    V : ndarray
        An ndarray of vectors of the velocity in the NEC frame.
    V_VSC : ndarray
        An ndarray of vectors of the velocity in the VSC frame.
    pos_ltl : ndarray
        An ndarray of scalars of positions where local time is acounted for.
        Each position reference the same position as positions output.
        Note tht this is off compared to the parameter positions.
    inclination :
        An ndarray of scalars of the inclination of the aproximation 
        of the magnetic field, where the calculations
        are supposed to be applied at.
            
    Examples
    --------
    >>> from pyfac.fac import *  # doctest: +SKIP
    >>> fac_from_file()  # doctest: +ELLIPSIS, +SKIP
    (array([...]), array([[...]]), array([...]), array([...]), array([[...]]), array([[...]]), array([[...]]), array([...]))
    """
    input_data = sw.call(fetch_data, options, use_pop=False)
    input_data['time'] *= 0.001  # conversion to seconds
    return (single_sat_fac_full(input_data),
            input_data)


if __name__ == '__main__':
    # Test docstrings
    import doctest
    doctest.testmod()

    results, input_data = fac_from_file()
    lat, fac = results[1][:, 0], results[3]
    reference = sw.read_cdf(
        'otherData/SW_OPER_FACATMS_2F_20160101T000000_20160101T235959_0301.cdf',
        lat='Latitude',
        fac='FAC')
    plt.plot(reference['lat'], reference['FAC'], 'b')
    plt.plot(lat, fac, 'r:')
    # plt.show()



""" The safe_user module provides basic security for login information.

This module is used to save login credentials safely in an encrypted file
and still have it employed by through an API,
with possibility for user prompting for password to open said file.

It uses AES encryption with random initial vector saved in file,
with the message containing padding if it is short,
typically less than 4096 characters.
No guarentee is granted on the strength of the randomization.
There is given no certification on the security level of data 
stored by the use of this module, so it is mainly recommended 
to be used for data requiring only lower level security,
such as access to unimportant/semi-public websites.
If you need high secrutity storage, such as for credit card
information or state or trade secrets, it is hightly suggested
you consult a security proffesional on what would be adequat.

Guide
-----
This module is a functional library, with an onion-like abstraction structure.
There are in general functions to add and remove single layers,
or add or remove all they way up to a specific layer.
You can use this to provide a solution on the level you need.

Each layer has their own section, so you can go look at that,
for information on that specific layer.
The layers are: XML, Padding, Encryption, Boxing, Locking.
There is also a section called Other, which is on parts of the module,
which falls outside the specific layers.

XML
---
This is the base structure used for communicating data.
This can be considered the core of the onion,
and most of functions can work with other data structures.
The exceptions are fetch_login, and those with XML in their name.
    
Functions on this layer:
    build_default_XML
        Build a minimal XML structure.
    find_login_in_XML
        Extract login information from the XML structure.
        
Padding
-------
When we add the padding layer we extend a text,
such that it is a valid input for the encryption layer.
We also use it to ensure that messages are not short,
and provide extra entropy.
The padding layer contains the meta-data needed to undo the padding.
A message which has this layer applied to it is called
a padded message/text or padded for short.

The format used is as follows:
'Hxxx*paddedMessage              '
Where H is a hexadeximal number describing how many extra spaces are placed
at the end to reach a length divisible by 16.
The xxx describes how many characters to skip
between each character in the message, it is terminated by the '*'

Functions on this layer:
    format_padding
        A function that takes a text and formats it into a valid padding.
        You can use this as a minimal padding scheme.
    remove_padding
        Recover the plain text from the padded text.
    add_padding
        Use sdefault way to add padding.
    add_padding1, add_padding2, add_padding3
        Different implementations of padding.
    set_padding_defaults
        Used to change add padding defaults, so one can use a different
        algorithm or base even when using functions that hide these details.
    
Encryption
----------
When we add the encryption layer we apply AES encryption on a padded text.
This requires a block size of 128 bits, which means the padded text
must effectively have a length divisable by 16 (assuming 8 bit charcters).
The key used is constructed from cryptographic hashing, 
so it is possible to use arbitary length passwords.
AES encryption requires an initialization vector (IV),
which is assumed to be of 1 block length (128 bits or 16 characters).
Failure to satisfy these requirements may result in undocumented errors,
as the layers around this is build to ensure that they are satisfied.
A message which has this layer applied to it is called
eitger an encrypted message/text or cipher text or cipher for short.

Functions on this layer:
    encrypt
        applies the encryption layer to a padded text.
    decrypt
        removes the rencryption layer and recover a padded text.

Boxing
------
The boxing layer is concerned with creating a package
of an encrypted message and its initialization vector.
The boxing layer only comes in the form where it applies
multiple layers at once, and it is intended as a simple
to use outer interface, where you do not need to concern
yourself with more than what message you have,
and what password to encrypt it with.
This means that both the padding layer and encryption
layer are also applied together with the boxing layer.
A message which has this layer applied to it is called a boxed message/text.

Functions in this layer:
    box_content
        Take a message (before padding layer),
        and apply layers with the boxing layer as the outermost layer.
    unbox_content
        Take a boxed message and remove,
        where the innermost layer removed is the padding layer.

Locking
-------
The locking layer is the use of file abstraction,
where the content of those files may have been boxed on unboxed.
Just like the boxing layer it is intended as a finishing layer,
where instead of being applied to strings like the boxing layer,
it is applied to files instead.
A file which has this layer applied to it is called
a locked file or an encrypted file.

Functions in this layer:
    lock_file
        Take a file with some content, and overwrite it with
        a version of it itself where its content have been boxed.
    unlock_file
        Take a file with some boxed content, and overwrite it with
        a version of it itself where its content have been unboxed.
    fetch_login
        Look into a file with boxed xml based content,
        unbox the content and return information from the file.
        All without changing the file.
    build_file
        Construct xml content similar to build_default_XML,
        and write a locked file with that content.
        
Other
-----
_chain
    A function used to fcreater flatter generators.
prompt_login_data
    A function used to prompt username and password from a user at runtime.
prompt_password
    A function used to prompt password from a user at runtime.
    
Examples
--------
>>> from pyfac.safety import *  # doctest: +SKIP
>>> c = box_content('One Ring to rule them all, '
...                 'One ring to find them; '
...                 'One ring to bring them all '
...                 'and in the darkness bind them.', password='Tolkien')
>>> len(c)
3456
>>> s = unbox_content(c,password='Tolkien')
>>> s[:26]
'One Ring to rule them all,'

>>> build_file('test.xml',
...             password='bukke bruse', 
...             default_login=('user1','failed'))  # doctest: +SKIP
>>> fetch_login('test.xml', password='bukke bruse')  # doctest: +SKIP
{'username': 'user1', 'password': 'failed')
"""

__version__ = '0.4.0'
__author__ = 'Ask Neve Gamby'

import random
import math
import string
import os.path

import hashlib
import getpass
import xml.etree.ElementTree as ET
from Crypto.Cipher import AES


def _chain(*args):
    if len(args) == 1:
        args = args[0]
    for arg in args:
        yield from arg


def prompt_login_data():
    """ Prompt a username and a password from the user.
    
    This function prompts the user two times,
    first for a username, secondly for a password.
    The password prompt hides the input from view,
    such that one can enter it 
    while others might be looking at the screen.
    
    Returns
    -------
    A tuple containing the following:
    user : str
        A string corrinsponding to the entered username.
    pw : str
        A string corrinsponding to the entered password.
        
    Note
    ----
    The input from the user is not sanity-checked,
    so it can form a popential vurnerability in deployed software.
    
    Examples
    --------
    >>> from pyfac.safety import *  # doctest: +SKIP
    >>> prompt_login_data() # doctest: +ELLIPSIS, +SKIP
    Enter login username:...
    Enter login password:...
    ('...', '...')
    """
    # """ Prompts the user for username and password,
    # and returns them.
    # Note that the password is typed in hidden.
    # """
    user = input('Enter login username:')
    pw = getpass.getpass('Enter login password:')
    return (user, pw)


def prompt_password(default='Def8lttPa88word',
                    is_bad=lambda x: len(x) < 1):
    """ Prompts the user for a password.
    
    This function prompts the user once for a password.
    The password prompt hides the input from view,
    such that one can enter it 
    while others might be looking at the screen.
    
    This function also allows some sanity checking of the
    input password, and will replace the password with
    a default if the users input fails the sanity check.
    
    Note that this function is intended to be for passwords,
    that represent the encryption keys to a file.
    
    Parameters
    ----------
    default : str
        A default password to use if the password given
        by the user does not pass the sanitycheck.
    is_bad : str -> boolean
        A function that takes a string (specifically a
        password candidate) and sanity chekcs it.
        Note that this function must return True
        if the candidate is to be discarded, and False
        if the candidate can be accepted.
        
    Returns
    -------
    str
        The password given by the user or the default password,
        if the password given by the user does not pass the sanity check.
        
    Examples
    --------
    >>> from pyfac.safety import *  # doctest: +SKIP
    >>> prompt_password() # doctest: +ELLIPSIS, +SKIP
    Enter password for file: ...
    ...
    """
    # """ Prompts the user for a password.
    # If the password is unsatisfactory it is replaced with a default.
    # """
    password = getpass.getpass('Enter password for file: ')
    if is_bad(password):
        password = default
    return password
    
    
def format_padding(text,stride=1,*args,**kwargs):
    """ Formats a message in relation with padding and its metadata.
    
    Format a message such that it includes the meta
    needed to remove the padding.
    It also may add whitespace at the end to ensure it can be AES encrypted.
    This function is sufficient for use of the encrypt function.
    
    The format looks as follows:
    'Hxxx*message    '
    H is a single hexadeximal digit, counting how many spaces
    have been added at the end.
    xxx is some digits representing stride in decimal character notation.
    There amount of digits in xxx is not fixed.
    The * specify when the xxx digits end and when the text starts.
    
    Parameters
    ----------
    text : str
        A string containing the message.
    stride : int, optional
        An integer describing the stride used for where in the text
        that the actual message can be located.
        Defaults to 1.
    *args
        Extra positional arguments, which are discared.
        This is here so it can be called with extra positional arguments.
    **kwargs
        Extra keyword arguments, which are discared.
        This is here so it can be called with extra keyword arguments.
        
    Returns
    -------
    str
        A string with meta-data. It satisfies len(s) % 16 == 0.
        
    Examples
    --------
    >>> from pyfac.safety import *  # doctest: +SKIP
    >>> format_padding('')
    'd0*             '
    >>> format_padding('short message')
    '00*short message'
    >>> format_padding('secret message')
    'f0*secret message               '
    >>> format_padding('DsReLcnrieBtJ 0mxeRsmsuaQgOe',2)
    '11*DsReLcnrieBtJ 0mxeRsmsuaQgOe '
    """
    # prefix = ['0',str(stride),'*']
    # if length is None:
        # length = len(text) + sum(map(len,prefix))
        
    text = str(stride-1) + '*' + text #we record skips instead of stride
    extra = (16 - ((len(text) + 1) % 16)) % 16
    return hex(extra)[2] + text + ' '*extra
    

    
def add_padding1(text, base=4096):
    """ Adds padding to the text so it fits well into encryption blocks.
    
    This function adds padding to a text, and include some metadata,
    which can be used to identify which parts are the padding,
    and thereby remove it.
    This function may add up to two types of padding,
    in addition to metadata that may also be added.
    The first padding type is a randomized padding that spreads
    out the input text to ensure that the text is not too short,
    and to provide resistance against length based attacks.
    The second padding type is blank padding used make the text
    compatible with block ciphers.
    This function is hardcoded to align to blocks of length 16.
    
    Warning
    -------
    This function has been depreciated, you should use add_padding2
    instead, since this one suffers performance problems.
    
    Parameters
    ----------
    text : str
        The text that is to be padded.
    base : int, optional
        An integer describing the boundary between short and long texts,
        where only short texts may be subject to the first type of padding.
        Default is 4096, must be at least 1.
        
    Returns
    -------
    str
        A string based on text, which has length % 16 == 0.
        It also contains the information needed to remove the padding.
        
    Examples
    --------
    >>> from pyfac.safety import *  # doctest: +SKIP
    >>> add_padding1('secret message') # doctest: +ELLIPSIS
    'b255*...s...e...c...r...e...t... ...m...e...s...s...a...g...e           '
    >>> add_padding1('secret message',16) # doctest: +ELLIPSIS
    'f0*secret message               '
    >>> add_padding1('secret message',8) # doctest: +ELLIPSIS
    'f0*secret message               '
    """
    # """ Adds enough padding to ensure it can easily be encrypted.
    # It also add random signal if the text is much shorter
    # than the base.
    # Warning: This function is not assymtothically inefficient
    # """
    padded = text
    applications = 0
    if len(text) < base:
        if len(padded) == 0:
            padded = ' '
            applications = 1
        while len(padded)<<1 <= base - len(str(base)) - 2:
            applications += 1
            temp = random.choices(
                string.ascii_uppercase
                + string.ascii_lowercase
                + string.digits,
                k=len(padded))
            padded = ''.join(k + c for (k, c) in zip(temp, padded))
    return format_padding(padded,1 << applications)
    # padded = str(1 << applications) + '*' + padded
    # extra = (16 - ((len(padded) + 1) % 16)) % 16
    # padded = hex(extra)[2] + padded + ' '*extra
    # return padded
    

def add_padding2(text, base=4096):
    """ Adds padding to the text so it fits well into encryption blocks.
    
    This function adds padding to a text, and include some metadata,
    which can be used to identify which parts are the padding,
    and thereby remove it.
    This function may add up to two types of padding,
    in addition to metadata that may also be added.
    The first padding type is a randomized padding that spreads
    out the input text to ensure that the text is not too short,
    and to provide resistance against length based attacks.
    The second padding type is blank padding used make the text
    compatible with block ciphers.
    This function is hardcoded to align to blocks of length 16.
    
    Parameters
    ----------
    text : str
        The text that is to be padded.
    base : int, optional
        An integer describing the boundary between short and long texts,
        where only short texts may be subject to the first type of padding.
        Default is 4096, must be at least 1.
        
    Returns
    -------
    str
        A string based on text, which has length % 16 == 0.
        It also contains the information needed to remove the padding.
        
    Examples
    --------
    >>> from pyfac.safety import *  # doctest: +SKIP
    >>> add_padding2('secret message') # doctest: +ELLIPSIS
    'b255*...s...e...c...r...e...t... ...m...e...s...s...a...g...e           '
    >>> add_padding2('secret message',16) # doctest: +ELLIPSIS
    'f0*secret message               '
    >>> add_padding2('secret message',8) # doctest: +ELLIPSIS
    'f0*secret message               '
    """
    # """ Adds enough padding to ensure it can easily be encrypted.
    # It also add random signal if the text is much shorter
    # than the base.
    # """
    padded = text
    # applications = 0
    if len(text) < base:
        applications = math.floor(math.log2(base/max(len(text), 1)))
        # print(applications,1<<applications,base)
        if len(text) == 0:
            padded = ' '
            # stride = 1<<(applications+1) 
        while len(padded) << applications > base - len(str(base))-2 and applications > 0:
            # print('applications: ' + str(applications) + ' -> ' + str(applications-1))
            applications -= 1
        stride = 1 << applications
        options = (string.ascii_uppercase
                   + string.ascii_lowercase
                   + string.digits)
        rand = random.SystemRandom()
        # print( applications, (1 << applications), (1 << applications) - 1)
        padded = ''.join(_chain(_chain(
            rand.choices(options, k=(1 << applications) - 1), (char,))
            for char in padded))
    else:
        stride = 1
    return format_padding(padded,stride if text else max(base+1,2))
    # padded = str(1 << applications) + '*' + padded
    # extra = (16 - ((len(padded) + 1) % 16)) % 16
    # # print(extra)
    # padded = hex(extra)[2] + padded + ' '*extra
    # return padded

    
def add_padding3(message, base=4096):
    stride = math.floor(base/max(len(message),1))
    stride = max(stride,1)
    # print(stride,stride-1)
    if len(message) < base:
        options = (string.ascii_uppercase
               + string.ascii_lowercase
               + string.digits)
        rand = random.SystemRandom()
        padded = ''.join(_chain(_chain(
            rand.choices(options, k=stride-1), (char,))
            for char in (message if message else ' ')))
    else:
        padded = message
    return format_padding(padded,stride if message else max(base+1,2))

    
def add_padding(message, base = 4096, algorithm=add_padding2):
    return algorithm(message,base=base)
    
    
def set_default_padding(base = 4096, algorithm=add_padding2):
    add_padding.__defaults__ = (base,algorithm)
    

def _split_padding(padded):
    """ Used for debugging."""
    extra = int(padded[0], 16)  # read first character as hexadeximal
    temp = padded[1:len(padded)-(extra)]
    skip, temp = temp.split('*', 1)
    stride = int(skip)+1
    if int(stride) > 0:
        result = temp[int(stride)-1::int(stride)]
        rand = ''.join(c for i,c in enumerate(temp) if i not in range(int(stride)-1,len(temp),int(stride)))
    return [extra, stride, '*', rand, result, padded[len(padded)-(extra):]]
    
    

def remove_padding(padded):
    """ Remove the padding from a string.
    
    This function removes the padding generated by an addpadding
    or the simpler format_padding, including the metadata.
    The text is assumed to satisfy the format given in format_padding.
    
    Parameters
    ----------
    padded : str
        A piece of text that have been padded by an add_padding.
        
    Returns
    -------
    str
        A string of the text without the padding.
        This should be equivalent to the text before padding was added.
        
    Examples
    --------
    >>> from pyfac.safety import *  # doctest: +SKIP
    >>> padded = add_padding('secret message')
    >>> remove_padding(padded)
    'secret message'
    >>> remove_padding(add_padding('other message',16))
    'other message'
    >>> remove_padding(add_padding('another message',8))
    'another message'
    """
    # base : int, optional
    # An integer describing the boundary between short and long texts,
    # where only short texts may be subject to the first type of padding.
    # It is important that this is the same as for the call to add_padding2,
    # as this data is not part of the metadata included in the padding.
    # Default is 4096, must be at least 1.
    # There is no guarentee for correct recovery of the message
    # and removal of padding, if the base used on add_padding2
    # and remove_padding are different, though for some cases
    # it will still work (such as if both have len(padded)> base).
    # """ removes the padding added by add_padding,
    # assuming they are done with the same base.
    # """
    extra = int(padded[0], 16)  # read first character as hexadeximal
    result = padded[1:len(padded)-(extra)]
    # if len(result) < base:
    skip, result = result.split('*', 1)
    stride = int(skip)+1
    if int(stride) > 0:
        result = result[int(stride)-1::int(stride)]
    return result


def encrypt(padded, password=None, IV=16 * '\x7A'):
    """ Encrypt a text with AES encryption.
    
    Encrypts a text with the AES encryption, with given intial vector,
    and a password either passed or prompted to the user at runtime.
    
    Parameters
    ----------
    padded : str
        A string of (padded) text to be encrypted.
        It must satisfy len(padded)%16 == 0.
    password : str, optional
        The password used to construct the key for the encryption.
        If it is not specified as a parameter,
        then the user will be prompted to choose one during runtime.
    IV : str, optional
        A string representing the initial vector of the AES encryption.
        Must be exactly 16 bytes long (equivalent to 16 normal characters).
        Default is 16 repeat '\x7A' == 'z' letters.
        
    Returns
    -------
    binary string
        A binary string containing the encrypted text.
    
    Examples
    --------
    >>> from pyfac.safety import *  # doctest: +SKIP
    >>> payload = add_padding('message',8)
    >>> encrypt(payload, password='ESA') # doctest: +ELLIPSIS
    b'...'
    """
    # """ Encrypt text with AES encryption,
    # with given password and InitialVector.
    # If no password is given the user will be promted for one.
    # Returns a binary string
    # """
    if password is None:
        password = prompt_password()
    hashed = hashlib.sha256(password.encode('utf-8')).digest()
    return AES.new(hashed, AES.MODE_CBC, IV=IV).encrypt(padded)


def decrypt(cipher_text, password=None, IV=16 * '\x7A'):
    """ Decrypt an AES encrypted message.
    
    Decrypts a text with the AES encryption, with given intial vector,
    and a password either passed or prompted to the user at runtime.
    
    Parameters
    ----------
    cipher_text : binary str
        A string of binary type, which contains an AES encrypted message.
    password : str, optional
        The password used to construct the key for the decryption.
        If it is not specified as a parameter,
        then the user will be prompted to choose one during runtime.
    IV : str, optional
        A string representing the initial vector of the AES encryption.
        Must be exactly 16 bytes long (equivalent to 16 normal characters).
        Default is 16 repeat '\x7A' == 'z' letters.
        
    Returns
    -------
    str
        A string representing the decrypted text.
        
    Examples
    --------
    >>> from pyfac.safety import *  # doctest: +SKIP
    >>> payload = add_padding('message',8)
    >>> cipher_text = encrypt(payload, password='ESA') # doctest: +ELLIPSIS
    >>> decrypt(cipher_text, password='ESA') == payload
    True
    """
    # """ Decrypt text with AES encryption,
    # with given password and InitialVector.
    # If no password is given the user will be promted for one.
    # """
    if password is None:
        password = prompt_password()
    hashed = hashlib.sha256(password.encode('utf-8')).digest()
    return AES.new(hashed, AES.MODE_CBC, IV=IV).decrypt(cipher_text).decode('utf-8')


def find_login_in_XML(text, target='viresLogin'):
    """ Search an XML string for login-like information.
    
    Search a string of xml structure of a specific form.
    It is assumed that the xml structure contains the following structure:
    Outermost a root of arbitary tag.
    Inside the root one or more taged pieces, one being "default"
    Inside these tags a set of tags with text content.
    
    Parameters
    ----------
    text : str
        A string containing a xml structure as described above.
    target : str, optional
        A string describing the tag to search for.
        Searches for 'default' if not found.
        Defaults to 'viresLogin'.
        
    Returns
    -------
    dict
        A dictionary containing the pairs of tags and text,
        for tags found as children to the search result.
        
    Examples
    --------
    >>> from pyfac.safety import *  # doctest: +SKIP
    >>> find_login_in_XML('<xml>'
    ...                   '  <default>'
    ...                   '    <username>admin</username>'
    ...                   '    <password>123</password>'
    ...                   '  </default>'
    ...                   '</xml>')
    {'username': 'admin', 'password': '123'}
    >>> find_login_in_XML('<xml>'
    ...                   '  <viresLogin>'
    ...                   '    <username>admin</username>'
    ...                   '    <password>123</password>'
    ...                   '  </viresLogin>'
    ...                   '</xml>')
    {'username': 'admin', 'password': '123'}
    >>> find_login_in_XML('<xml>'
    ...                   '  <default>'
    ...                   '    <username>admin</username>'
    ...                   '    <password>123</password>'
    ...                   '  </default>'
    ...                   '  <viresLogin>'
    ...                   '    <username>John</username>'
    ...                   '    <password>Doe</password>'
    ...                   '  </viresLogin>'
    ...                   '</xml>')
    {'username': 'John', 'password': 'Doe'}
    >>> find_login_in_XML('<xml>'
    ...                   '  <default>'
    ...                   '    <username>admin</username>'
    ...                   '    <password>123</password>'
    ...                   '  </default>'
    ...                   '  <viresLogin>'
    ...                   '    <username>John</username>'
    ...                   '    <password>Doe</password>'
    ...                   '  </viresLogin>'
    ...                   '  <trouble>'
    ...                   '    <username>snowflake</username>'
    ...                   '    <password>unique</password>'
    ...                   '  </trouble>'
    ...                   '</xml>',
    ...                   'trouble')
    {'username': 'snowflake', 'password': 'unique'}
    >>> find_login_in_XML('<xml>'
    ...                   '  <default>'
    ...                   '    <username>admin</username>'
    ...                   '    <password>123</password>'
    ...                   '  </default>'
    ...                   '  <viresLogin>'
    ...                   '    <username>John</username>'
    ...                   '    <password>Doe</password>'
    ...                   '  </viresLogin>'
    ...                   '  <trouble>'
    ...                   '    <username>snowflake</username>'
    ...                   '    <password>unique</password>'
    ...                   '  </trouble>'
    ...                   '</xml>',
    ...                   'globus')
    {'username': 'admin', 'password': '123'}
    
    Note
    ----
    Using build_file will generate a valid xml structure,
    containing only the default tag and subtags.
    """
    # """ This will find the login information under target
    # in the given xml text.
    # If target is not found, then default may be used instead.
    # """
    root = ET.fromstring(text)
    place = root.find(target)
    if place is None:
        place = root.find('default')
    return {elem.tag: elem.text for elem in place}
    

def build_default_XML(default_login=None):
    """ Build a xml based string with default login settings.
    
    This function creates a string of XML, which contains
    default login settings, which can then be read
    by find_login_in_XML or saved to a file.
    It is recommened to have this information encrypted,
    such as by using box_content on it.
    
    Parameters
    ----------
    default_login : tuple of two strings, optional
        A tuple of form (username, password), which is used
        as the default login informaion recorded in the xml string.
        If it is not specified as a parameter,
        then the user will be prompted to choose one during runtime.
    
    Returns
    -------
    str
        A string representing an xml file, which contains
        default login settings and would work for `find_login_in_XML`.
        
    Examples
    --------
    >>> from pyfac.safety import *  # doctest: +SKIP
    >>> build_default_XML(('nimda', 'tapas')) # doctest: +ELLIPSIS
    '<?xml version="1.0"?>...<xml>...<default>...<username>nimda</username>...<password>tapas</password>...</default>...</xml>...'
    """
    if default_login is None:
        (user, password) = prompt_login_data()
    else:
        (user, password) = default_login
    content = """<?xml version="1.0"?>
<xml>
    <default>
        <username>""" + user + """</username>
        <password>""" + password + """</password>
    </default>
</xml>
"""
    return content  # Easier to see where the function ends this way.


def box_content(text, password=None):
    """ Package text into encrypted cypther-text with meta-data.
    
    This function 'box' the text as an AES encrympted dump.
    It takes care of all the needed substeps, including
    padding, choosing initial vector, encryption, and storing meta-data.
    
    Parameters
    ----------
    text : str
        A string of plain text.
    password : str, optional
        The password used to construct the key for the encryption.
        If it is not specified as a parameter,
        then the user will be prompted to choose one during runtime.
    
    Returns
    -------
    str
        A binary string containing the encrypted text, initial vector,
        and other meta-data which is hidden under the encryption.
        
    Examples
    --------
    >>> from pyfac.safety import *  # doctest: +SKIP
    >>> box_content('There is no cake', 'beware of goats') # doctest: +ELLIPSIS
    b'...'
    """
    # """ The text is 'boxed' as a encrypted dump.
    # boxing includes padding the message and including
    # the meta-data needed to undo the process.
    # If password is not given the user will be prompted for one.
    # """
    IV = ''.join(str(random.choice(string.hexdigits)) for __ in range(16))
    boxed = encrypt(add_padding(text), password, IV)
    return IV.encode('utf-8') + boxed


def unbox_content(boxed, password=None):
    """ Retrive text from cyper-text from box_content.
    
    This function takes some text that have been 'boxed'
    as an AES encrypted dump, and revert back into the original test.
    It takes care of all the needed substeps, including
    removing padding, extracting meta-data and decryption.
    
    Parameters
    ----------
    boxed : str
        A string of binary text, representing a result from box_content.
    password : str, optional
        The password used to construct the key for the decryption.
        If it is not specified as a parameter,
        then the user will be prompted to choose one during runtime.
    
    Returns
    -------
    str
        A string representing the decrypted and unpadded text.
        
    Examples
    --------
    >>> from pyfac.safety import *  # doctest: +SKIP
    >>> cipher = box_content('send the rover', 'once in a blue moon')
    >>> unbox_content(cipher, 'once in a blue moon')
    'send the rover'
    """
    # """ Retrive the content of a 'boxed' text.
    # This is functionally the inverse of box_content.
    # If password is not given the user will be prompted for one.
    # """
    decrypted = decrypt(boxed[16:], password, boxed[0:16])
    return remove_padding(decrypted)


def lock_file(filename, password=None):
    """ Replace a file with a boxed version of its content.
    
    This function reads the file at the path `filename`,
    box the content and write the result back over the file.
    
    Parameters
    ----------
    filename : str
        A string of the path to the file, including its name.
    password : str, option
        The password used to construct the key for the encryption.
        If it is not specified as a parameter,
        then the user will be prompted to choose one during runtime.
        
    Returns
    -------
    None
        There is no return statement, this results in a None.
        
    Examples
    --------
    >>> from pyfac.safety import *  # doctest: +SKIP
    >>> lock_file('some.xml','gatekeeper') # doctest: +SKIP
    """
    # """ Replace a file of filename with
    # the boxed version of its content.
    # If password is not given the user will be prompted for one.
    # """
    with open(filename, 'r') as f:
        content = f.read()
    out = box_content(content, password)
    with open(filename, 'wb') as f:
        f.write(out)


def build_file(filename, password=None, default_login=None):
    """ Construct a locked xml file which include a default loging.
    
    This function is used to construct a defaulted boxed xml file.
    This can be used with fetch_loging, or unboxed with unlock file.
    The content of the file, after unboxing can be used with
    find_login_in_XML.
    
    Parameters
    ----------
    filename : str
        A string of the path to the file, including its name.
    password : str, optional
        The password used to construct the key for the encryption.
        If it is not specified as a parameter,
        then the user will be prompted to choose one during runtime.
    default_login : tuple of two strings, optional
        A tuple of form (username, password), which is used
        as the default login informaion recorded in the xml file.
        If it is not specified as a parameter,
        then the user will be prompted to choose one during runtime.
        
    Returns
    -------
    None
        There is no return statement, this results in a None.
        
    Examples
    --------
    >>> from pyfac.safety import *  # doctest: +SKIP
    >>> build_file('example.xml', 'welcome', ('blank', 'stare')) # doctest: +SKIP
    """
    """ Construct a boxed (encrypted with metadata) file
    of filename, the content will include default login
    information, which the user will be prompted for.
    If password for the file is not given the user will be prompted for one.
    """
    # if default_login is None:
        # (user, password) = prompt_login_data()
    # else:
        # (user, password) = default_login
    # content = """<?xml version="1.0"?>
# <xml>
    # <default>
        # <username>""" + user + """</username>
        # <password>""" + password + """</password>
    # </default>
# </xml>
# """
    out = box_content(build_default_XML(default_login))
    with open(filename, 'wb') as f:
        f.write(out)


def fetch_login(filename, target='viresLogin', password=None):
    """ Fetch information from a locked file.
    
    This function peaks into an encrypted file and retrievs
    login like information from it, without leaving the file unencrypted.
    Specifically it uses files that have been locked,
    meaning it has boxed content.
    It is assumed that the content after unboxing will form a valid
    xml structure, that find_login_in_XML can be applied to.
    
    This function is intended to be used as a top level abstraction,
    of aquiring login information in other modules.
    
    Parameters
    ----------
    filename : str
        A string of the path to the file, including its name.
    target : str, optional
        A string representing the xml tag where the data
        we want is stored under.
        If this tag does not exists we will get the data stored
        uder the tag 'default' instead.
        Default is 'viresLogin'.
    password : str, optional
        The password used to construct the key for the decryption.
        If it is not specified as a parameter,
        then the user will be prompted to choose one during runtime.
        
    Returns
    -------
    None
        There is no return statement, this results in a None.
    
    Examples
    --------
    >>> from pyfac.safety import *  # doctest: +SKIP
    >>> fetch_login('example.xml', 'welcome') # doctest: +SKIP
    {'username': 'blank', 'password': 'stare'}
    """
    # """ Peek into an encrypted file of filename,
    # and extract login information on the given target (with default).
    # If password for the file is not given the user will be prompted for one.
    # If the file does not exist then a valid file may be constructed
    # based on input from the user.

    # Note that this is the main functionally expected to be used outside,
    # since it can buildup the necessary structure
    # by just including this function,
    # and not manually having to set up the rest.
    # """
    if not os.path.isfile(filename):
        print('The target file: ' + filename
              + ' do not exist yet, constructing it now...')
        build_file(filename, password)
    with open(filename, 'rb') as f:
        text = f.read()
    return find_login_in_XML(unbox_content(text, password), target)


def unlock_file(filename, password=None):
    """ Replace a file with an unboxed version of its content.

    This function reads the file at the path `filename`,
    unbox the content and write the result back over the file.
    
    Parameters
    ----------
    filename : str
        A string of the path to the file, including its name.
    password : str, option
        The password used to construct the key for the decryption.
        If it is not specified as a parameter,
        then the user will be prompted to choose one during runtime.
        
    Returns
    -------
    None
        There is no return statement, this results in a None.
        
    Examples
    --------
    >>> from pyfac.safety import *  # doctest: +SKIP
    >>> unlock_file('some.xml','gatekeeper') # doctest: +SKIP
    """
    # """ Replace a file of filename with
    # the unboxed version of its content.
    # If password is not given the user will be prompted for one.
    # """
    with open(filename, 'rb') as f:
        content = f.read()
    out = unbox_content(content, password)
    with open(filename, 'w') as f:
        f.write(out)


if __name__ == '__main__':
    # Test docstrings
    import doctest
    doctest.testmod()
